# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2021
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

import logging
import os
import sarracenia
import ssl
import subprocess
import sys

from sarracenia.transfer import Transfer
from sarracenia.transfer import alarm_cancel, alarm_set, alarm_raise

import urllib.error, urllib.parse, urllib.request
from urllib.parse import unquote

logger = logging.getLogger(__name__)


class Https(Transfer):
    """
    HyperText Transfer Protocol (HTTP)  ( https://en.wikipedia.org/wiki/Hypertext_Transfer_Protocol ) 
    sarracenia transfer protocol subclass supports/uses additional custom options:

    * accelWgetCommand (default: '/usr/bin/wget %s -o - -O %d' )

    built with: 
         urllib.request ( https://docs.python.org/3/library/urllib.request.html )
    """
    def __init__(self, proto, options):

        super().__init__(proto, options)

        self.o.add_option('accelWgetCommand', 'str', '/usr/bin/wget %s -o - -O %d')

        logger.debug("sr_http __init__")

        self.tlsctx = ssl.create_default_context()
        if hasattr(self.o, 'tlsRigour'):
            self.o.tlsRigour = self.o.tlsRigour.lower()
            if self.o.tlsRigour == 'lax':
                self.tlsctx = ssl.create_default_context()
                self.tlsctx.check_hostname = False
                self.tlsctx.verify_mode = ssl.CERT_NONE

            elif self.o.tlsRigour == 'strict':
                self.tlsctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
                self.tlsctx.options |= ssl.OP_NO_TLSv1
                self.tlsctx.options |= ssl.OP_NO_TLSv1_1
                self.tlsctx.check_hostname = True
                self.tlsctx.verify_mode = ssl.CERT_REQUIRED
                self.tlsctx.load_default_certs()
                # TODO Find a way to reintroduce certificate revocation (CRL) in the future
                #  self.tlsctx.verify_flags = ssl.VERIFY_CRL_CHECK_CHAIN
                #  https://github.com/MetPX/sarracenia/issues/330
            elif self.o.tlsRigour == 'normal':
                pass
            else:
                logger.warning(
                    "option tlsRigour must be one of:  lax, normal, strict")

        self.init()

    def registered_as():
        return ['http', 'https']

    # cd
    def cd(self, path):
        logger.debug("sr_http cd %s" % path)
        self.cwd = os.path.dirname(path)
        self.path = path

    # for compatibility... always new connection with http
    def check_is_connected(self):
        logger.debug("sr_http check_is_connected")

        if not self.connected : return False

        if self.sendTo != self.o.sendTo:
            self.close()
            return False

        return True

    # close
    def close(self):
        logger.debug("sr_http close")
        self.init()

    # connect...
    def connect(self):
        logger.debug("sr_http connect %s" % self.o.sendTo)

        if self.connected: self.close()

        self.connected = False
        self.sendTo = self.o.sendTo
        self.timeout = self.o.timeout

        if not self.credentials(): return False

        return True

    # credentials...
    def credentials(self):
        logger.debug("sr_http credentials %s" % self.sendTo)

        try:
            ok, details = self.o.credentials.get(self.sendTo)
            if details: url = details.url

            self.user = url.username if url.username != '' else None
            self.password = url.password if url.password != '' else None
            self.bearer_token = details.bearer_token if hasattr(
                details, 'bearer_token') else None

            return True

        except:
            logger.error(
                "sr_http/credentials: unable to get credentials for %s" %
                self.sendTo)
            logger.debug('Exception details: ', exc_info=True)

        return False

    # get
    def get(self,
            msg,
            remote_file,
            local_file,
            remote_offset=0,
            local_offset=0,
            length=0):
        logger.debug("get %s %s %d" % (remote_file, local_file, local_offset))
        logger.debug("sr_http self.path %s" % self.path)

        # open self.http

        if 'retrievePath' in msg:
            url = self.sendTo + '/' + msg['retrievePath']
        else:
            u = urllib.parse.urlparse( self.sendTo )
            url = u.scheme + '://' + u.netloc + '/' + urllib.parse.quote(self.path + '/' +
                                                              remote_file)

        ok = self.__open__(url, remote_offset, length)

        if not ok: return False

        # read from self.http write to local_file

        rw_length = self.read_writelocal(remote_file, self.http, local_file,
                                         local_offset, length)

        return rw_length

    def getAccelerated(self, msg, remote_file, local_file, length):

        arg1 = msg['baseUrl'] + '/' + msg['relPath']
        arg1 = arg1.replace(' ', '\ ')
        arg2 = local_file

        cmd = self.o.accelWgetCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()
        logger.info("accel_wget: %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            logger.warning("binary accelerator %s returned: %d" % ( cmd, p.returncode ) )
            return -1
        # FIXME: length is not validated.
        return length

    # init
    def init(self):
        Transfer.init(self)

        logger.debug("sr_http init")
        self.connected = False
        self.http = None
        self.details = None
        self.seek = True

        self.urlstr = ''
        self.path = ''
        self.cwd = ''

        self.data = ''
        self.entries = {}


# ls

    def ls(self):
        logger.debug("sr_http ls")

        # open self.http

        self.entries = {}

        url = self.sendTo + '/' + urllib.parse.quote(self.path)

        ok = self.__open__(url)

        if not ok: return self.entries

        # get html page for directory

        try:
            dbuf = None
            while True:
                alarm_set(self.o.timeout)
                chunk = self.http.read(self.o.bufsize)
                alarm_cancel()
                if not chunk: break
                if dbuf: dbuf += chunk
                else: dbuf = chunk

            #self.data = dbuf.decode('utf-8')

            # invoke option defined on_html_page ... if any

            #for plugin in self.o.on_html_page_list:
            #    if not plugin(self):
            #        logger.warning("something wrong")
            #        return self.entries

        except:
            logger.warning("sr_http/ls: unable to open %s" % self.urlstr)
            logger.debug('Exception details: ', exc_info=True)

        return dbuf

    # open
    def __open__(self, path, remote_offset=0, length=0):
        logger.debug( f"{path}")

        self.http = None
        self.connected = False
        self.req = None
        self.urlstr = path

        # have noticed that some site does not allow // in path
        if path.startswith('http://') and '//' in path[7:]:
            self.urlstr = 'http://' + path[7:].replace('//', '/')

        if path.startswith('https://') and '//' in path[8:]:
            self.urlstr = 'https://' + path[8:].replace('//', '/')

        alarm_set(self.o.timeout)

        try:
            # when credentials are needed.
            headers = {'user-agent': 'Sarracenia ' + sarracenia.__version__}
            if self.bearer_token:
                logger.debug('bearer_token: %s' % self.bearer_token)
                headers['Authorization'] = 'Bearer ' + self.bearer_token

            if self.user != None:
                password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
                # takeaway credentials info from urlstr
                cred = self.user + '@'
                self.urlstr = self.urlstr.replace(cred, '')
                if self.password != None:
                    cred = self.user + ':' + self.password + '@'
                    self.urlstr = self.urlstr.replace(cred, '')

                # continue with authentication
                password_mgr.add_password(None, self.urlstr, self.user,
                                          unquote(self.password))
                auth_handler = urllib.request.HTTPBasicAuthHandler(
                    password_mgr)

                #hctx = ssl.create_default_context()
                #hctx.check_hostname = False
                #hctx.verify_mode = ssl.CERT_NONE
                ssl_handler = urllib.request.HTTPSHandler(0, self.tlsctx)

                # create "opener" (OpenerDirector instance)
                opener = urllib.request.build_opener(auth_handler, ssl_handler)

                # use the opener to fetch a URL
                opener.open(self.urlstr)

                # Install the opener.
                urllib.request.install_opener(opener)

            # Now all calls to get the request use our opener.
            self.req = urllib.request.Request(self.urlstr, headers=headers)

            # set range in byte if needed
            if remote_offset != 0:
                str_range = 'bytes=%d-%d' % (remote_offset,
                                             remote_offset + length - 1)
                self.req.headers['Range'] = str_range

            # https without user : create/use an ssl context
            ctx = None
            if self.user == None and self.urlstr.startswith('https'):
                ctx = self.tlsctx
                #ctx.check_hostname = False
                #ctx.verify_mode = ssl.CERT_NONE

            # open... we are connected
            if self.timeout == None:
                self.http = urllib.request.urlopen(self.req, context=ctx)
            else:
                self.http = urllib.request.urlopen(self.req,
                                                   timeout=self.timeout,
                                                   context=ctx)

            self.connected = True

            alarm_cancel()

            return True

        except urllib.error.HTTPError as e:
            logger.error('Download failed 4 %s ' % self.urlstr)
            logger.error(
                'Server couldn\'t fulfill the request. Error code: %s, %s' %
                (e.code, e.reason))
            alarm_cancel()
            self.connected = False
            raise
        except urllib.error.URLError as e:
            logger.error('Download failed 5 %s ' % self.urlstr)
            logger.error('Failed to reach server. Reason: %s' % e.reason)
            alarm_cancel()
            self.connected = False
            raise
        except:
            logger.warning("unable to open %s" % self.urlstr)
            logger.debug('Exception details: ', exc_info=True)
            self.connected = False
            alarm_cancel()
            raise

        alarm_cancel()
        return False



"""
Posts updated files of USGS hydrometric data. Station site IDs provided
in the poll_usgs_stn_file. Compatible with Python 3.5+.

Status: UNTESTED... don't have a working basis (even on v2) to compare against.
        was updated for v3. but don't have credentials to confirm it now works.

usage:
	in an sr3 poll configuration file:
	pollUrl http://waterservices.usgs.gov/nwis/iv/?format=waterml,2.0&indent=on&site={0:}&period=PT3H&parameterCd=00060,00065,00011

	batch [station_chunk]
	poll_usgs_station  station-declaration
	callback gather/usgs

	If multiple usgs stations need to be fetched in one call, station_chunk should specify how big the station
	blocks should be. If not set it'll individually download station data.
	If poll_usgs_stn_file isn't set, it'll default to pulling the siteIDs from: 
	https://water.usgs.gov/osw/hcdn-2009/HCDN-2009_Station_Info.xlsx
	directory. each station declaration is in the following format:
	SourceID | SiteID | SiteCode | SiteName | CountryID | StateID | UTCOffset
	Each station on its own line.

        to reformat into a parseable configuration option, declare poll_usgs_station as a list option,
        so each line looks like::

             poll_usgs_station 7|70026|9014087|Dry Dock, MI|US|MI|-5.0

        then this file can be an *include* directive, and include supports both local and remote url's
        so the db can be admined in some other http accessible git repository.

	More info on the http rest parameters at: https://waterservices.usgs.gov/rest/IV-Service.html
	For writing fault-resistant code that polls from usgs: https://waterservices.usgs.gov/docs/portable_code.html
	Sign up for updates involving if/how the format changes: http://waterdata.usgs.gov/nwis/subscribe?form=email
	Parameter codes to tailor the data you want: 
	https://help.waterdata.usgs.gov/code/parameter_cd_query?fmt=rdb&inline=true&group_cd=%
	Currently the parametercd is set to:
	00060	Physical	Discharge, cubic feet per second
	00065	Physical	Gage height, feet
	00011	Physical	Temperature, water, degrees Fahrenheit	
"""

import datetime
import logging
import pandas as pd
import paramiko
import sarracenia
from sarracenia.flowcb import FlowCB
import urllib.request

logger = logging.getLogger(__name__)


class Usgs(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)
        self.o.add_option('poll_usgs_station', 'list')

        # Parse sitecodes from file if provided, or the usgs website (turns excel spreadsheet into pandas
        # dataframe, parses from there)
        self.sitecodes = []
        if hasattr(self.o, 'poll_usgs_station'):
            for s in self.o.poll_usgs_station:
                items = s.split('|')
                self.sitecodes.append(items[2])
            logger.info('%d stations declared' % len(self.sitecodes))
        else:
            df = pd.read_excel(
                'https://water.usgs.gov/osw/hcdn-2009/HCDN-2009_Station_Info.xlsx'
            )
            for row in df.iterrows():
                self.sitecodes.append(str(row[1]['STATION ID']).zfill(8))

        if hasattr(self.o, 'batch'):
            mult = True
            chunk_size = int(self.o.batch)

    def poll(self):

        run_time = datetime.datetime.utcnow().strftime('%Y%m%d_%H%M')

        gathered_messages = []
        if self.o.batch > 1:
            file_cnt = 0
            for sites in [
                    self.sitecodes[i:i + self.o.batch]
                    for i in range(0, len(self.sitecodes), self.o.batch)
            ]:
                stns = ','.join([s for s in sites])
                file_cnt += 1
                logger.debug('getting: %s' % self.o.pollUrl.format(stns))

                status_code = urllib.request.urlopen(
                    self.o.pollUrl.format(stns)).getcode()
                if status_code == 200:
                    logger.info("poll_usgs file updated %s" %
                                self.o.pollUrl.format(stns))

                    self.o.msg.new_baseurl = self.o.pollUrl.format(stns)

                    m = sarracenia.Message.fromFileInfo(
                        'usgs_{0}_sites{1}.xml'.format(run_time, file_cnt),
                        self.o)
                    gathered_messages.append(m)
                elif status_code == 403:
                    logger.error(
                        '''poll_usgs: USGS has determined your usage is excessive and \ 
							blocked your IP. Use the contact form on their site to be \
							unblocked.''')
                else:
                    logger.debug("poll_usgs file not found: %s" %
                                 self.o.pollUrl.format(stns))
        else:  # Get stations one at a time
            for site in self.sitecodes:
                logger.debug('getting: %s' % self.o.pollUrl.format(site))
                status_code = urllib.request.urlopen(
                    self.o.pollUrl.format(site)).getcode()
                if status_code == 200:
                    logger.info("poll_usgs file updated %s" %
                                self.o.pollUrl.format(site))
                    self.o.msg.new_baseurl = self.o.pollUrl.format(site)
                    m = sarracenia.Message.fromFileInfo(
                        'usgs_{0}_{1}.xml'.format(run_time, site), self.o)
                    gathered_messages.append(m)
                elif status_code == 403:
                    logger.error(
                        '''poll_usgs: USGS has determined your usage is excessive and \
							blocked your IP. Use the contact form on their site to be \
							unblocked.''')
                else:
                    logger.debug("poll_usgs file not found: %s" %
                                 self.o.pollUrl.format(site))
        return gathered_messages



#
# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# Questions or bugs report: dps-client@ec.gc.ca
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
# __init__.py : contains version number of sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#
from ._version import __version__


from base64 import b64decode, b64encode
import calendar
import datetime
import importlib.util
import logging
import os
import os.path
import random
import re
from sarracenia.featuredetection import features
import stat as os_stat
import sys
import time
import types
import urllib
import urllib.request

logger = logging.getLogger(__name__)

def baseUrlParse( url ):
    upr = urllib.parse.urlparse(url)
    u = types.SimpleNamespace()
    u.scheme = upr.scheme
    u.netlog = upr.netloc
    u.params = upr.params
    u.query = upr.query
    u.fragment = upr.fragment
    u.path = upr.path
    if u.scheme in [ 'sftp', 'file' ]:
        while u.path.startswith('//'):
            u.path = u.path[1:]
    return u


if features['filetypes']['present']:
   import magic

if features['mqtt']['present']:
   import paho.mqtt.client
   if not hasattr( paho.mqtt.client, 'MQTTv5' ):
       # without v5 support, mqtt is not useful.
       features['mqtt']['present'] = False

# if humanize is not present, compensate...
if features['humanize']['present']:
    import humanize

    def naturalSize( num ):
        return humanize.naturalsize(num,binary=True)

    def naturalTime( dur ):
        return humanize.naturaltime(dur)

else:
  
    def naturalSize( num ):
       return "%g" % num

    def naturalTime( dur ):
       return "%g" % dur


if features['appdirs']['present']:
    import appdirs

    def site_config_dir( app, author ):
        return appdirs.site_config_dir( app, author )

    def user_config_dir( app, author ):
        return appdirs.user_config_dir( app, author )

    def user_cache_dir( app, author ):
        return appdirs.user_cache_dir( app, author )
else:
    # if appdirs is missing, pretend we're on Linux.
    import pathlib

    def site_config_dir( app, author ):
        return '/etc/xdg/xdg-ubuntu-xorg/%s' % app

    def user_config_dir( app, author ):
        return str(pathlib.Path.home()) + '/.config/%s' % app
 
    def user_cache_dir( app, author ):
        return str(pathlib.Path.home()) + '/.cache/%s' % app

"""
 end of extra feature scan. 

"""

import sarracenia.filemetadata

class Sarracenia:
    """
        Core utilities of Sarracenia. The main class here is sarracenia.Message.
        a Sarracenia.Message is subclassed from a dict, so for most uses, it works like the 
        python built-in, but also we have a few major entry points some factoryies:
    

        **Building a message from a file**

        m = sarracenia.Message.fromFileData( path, options, lstat )
     
        builds a notification message from a given existing file, consulting *options*, a parsed
        in memory version of the configuration settings that are applicable

        **Options**

        see the sarracenia.config.Config class for functions to parse configuration files
        and create corresponding python option dictionaries. One can supply small 
        dictionaries for example::

          options['topicPrefix'] = [ 'v02', 'post' ]
          options['bindings'] = [ ('xpublic', [ 'v02', 'post'] , [ '#' ] )]
          options['queueName'] = 'q_anonymous_' + socket.getfqdn() + '_SomethingHelpfulToYou'

        Above is an example of a minimal options dictionary taken from the tutorial 
        example called moth_api_consumer.py. often 
        

        **If you don't have a file**

        If you don't have a local file, then build your notification message with:
    
        m = sarracenia.Message.fromFileInfo( path, options, lstat )
    
        where you can make up the lstat values to fill in some fields in the message.
        You can make a fake lstat structure to provide these values using sarracenia.filemetadata
        class which is either an alias for paramiko.SFTPAttributes 
        ( https://docs.paramiko.org/en/latest/api/sftp.html#paramiko.sftp_attr.SFTPAttributes )
        if paramiko is installed, or a simple emulation if not.
    
    
        from  sarracenia.filemetadata import FmdStat
    
        lstat = FmdStat()
        lstat.st_mtime= utcinteger second count in UTC (numeric version of a Sarracenia timestamp.)
        lstat.st_atime=
        lstat.st_mode=0o644
        lstat.st_size= size_in_bytes
    
        optional fields that may be of interest:
        lstat.filename= "nameOfTheFile"
        lstat.longname= 'lrwxrwxrwx    1 peter    peter          20 Oct 11 20:28 nameOfTheFile'
     
        that you can then provide as an *lstat* argument to the above *fromFileInfo()* 
        call. However the notification message returned will lack an identity checksum field.
        once you get the file, you can add the Identity field with:
    
        m.__computeIdentity(path, o):
    
        In terms of consuming notification messages, the fields in the dictionary provide metadata
        for the announced resource. The anounced data could be embedded in the notification message itself,
        or available by a URL.
    
        Messages are generally gathered from a source such as the Message Queueing Protocol wrapper
        class: moth... sarracenia.moth. 
    
    
        data = m.getContent()
    
        will return the content of the announced resource as raw data.
    
    """
    pass


class TimeConversions:
    """
    
     Time conversion routines.  

     * os.stat, and time.now() return floating point 

     * The floating point representation is a count of seconds since the beginning of the epoch.

     * beginning of epoch is platform dependent, and conversion to actual date is fraught (leap seconds, etc...)

     * Entire SR_* formats are text, no floats are sent over the protocol 
       (avoids byte order issues, null byte / encoding issues, and enhances readability.) 

     * str format: YYYYMMDDHHMMSS.msec goal of this representation is that a naive 
       conversion to floats yields comparable numbers.

     * but the number that results is not useful for anything else, so need these 
       special routines to get a proper epochal time.

     * also OK for year 2032 or whatever (rollover of time_t on 32 bits.)

     * string representation is forced to UTC timezone to avoid having to communicate timezone.
    
     timestr2flt() - accepts a string and returns a float.
    
     caveat

     FIXME: this encoding will break in the year 10000 (assumes four digit year) 
     and requires leading zeroes prior to 1000. One will have to add detection of 
     the decimal point, and change the offsets at that point.
        
    """
    pass


def stat( path ) -> sarracenia.filemetadata.FmdStat:
    """
       os.stat call replacement which improves on it by returning
       and SFTPAttributes structure, in place of the OS stat one,
       featuring:
 
       * mtime and ctime with subsecond accuracy 
       * fields that can be overridden (not immutable.)

    """
    native_stat = os.stat( path )
    
    sa = sarracenia.filemetadata.FmdStat()
    sa.st_mode = native_stat.st_mode
    sa.st_ino = native_stat.st_ino
    sa.st_dev  = native_stat.st_dev
    # st_nlink does not exist in paramiko.SFTPAttributes()
    #  FmdStat comes from that type.
    #sa.st_nlink  = native_stat.st_nlink
    sa.st_uid  = native_stat.st_uid
    sa.st_gid  = native_stat.st_gid
    sa.st_size  = native_stat.st_size

    sa.st_mtime = os.path.getmtime(path)
    sa.st_atime = os.path.getctime(path)
    sa.st_ctime = native_stat.st_atime
    return sa

def nowflt():
    return timestr2flt(nowstr())


def nowstr():
    return timeflt2str(time.time())


def timeflt2str(f=None):
    """
        timeflt2str - accepts a float and returns a string.

        flow is a floating point number such as returned by time.now()
        (number of seconds since beginning of epoch.)

        the str is YYYYMMDDTHHMMSS.sssss

        20210921T011331.0123

        translates to: Sept. 21st, 2021 at 01:13 and 31.0123 seconds.
        always UTC timezone.    
    """

    nsec = "{:.9g}".format(f % 1)[1:]
    return "{}{}".format(time.strftime("%Y%m%dT%H%M%S", time.gmtime(f)), nsec)


def timeValidate(s) -> bool:

    if len(s) < 14: return False
    if (len(s) > 14) and (s[8] != 'T') and (s[14] != '.'): return False
    if (len(s) > 15) and (s[8] == 'T') and (s[15] != '.'): return False
    if not s[0:8].isalnum(): return False
    return True


def timestr2flt(s):

    if s[8] == "T":
        s = s.replace('T', '')
    dt_tuple = int(s[0:4]), int(s[4:6]), int(s[6:8]), int(s[8:10]), int(
        s[10:12]), int(s[12:14])
    t = datetime.datetime(*dt_tuple, tzinfo=datetime.timezone.utc)
    return calendar.timegm(t.timetuple()) + float('0' + s[14:])


def timev2tov3str(s):
    if s[8] == 'T':
        return s
    else:
        return s[0:8] + 'T' + s[8:]


def durationToSeconds(str_value, default=None) -> float:
    """
   this function converts duration to seconds.
   str_value should be a number followed by a unit [s,m,h,d,w] ex. 1w, 4d, 12h
   return 0.0 for invalid string.
   """
    factor = 1

    if type(str_value) in [list]:
        str_value = str_value[0]

    if type(str_value) in [int, float]:
        return str_value
    
    if type(str_value) is not str:
        return 0

    if str_value.lower() in [ 'none', 'off', 'false' ]:
        return 0

    if default and str_value.lower() in [ 'on', 'true' ]:
        return float(default)

    if str_value[-1] in 'sS': factor *= 1
    elif str_value[-1] in 'mM': factor *= 60
    elif str_value[-1] in 'hH': factor *= 60 * 60
    elif str_value[-1] in 'dD': factor *= 60 * 60 * 24
    elif str_value[-1] in 'wW': factor *= 60 * 60 * 24 * 7
    if str_value[-1].isalpha(): str_value = str_value[:-1]

    try:
        duration = float(str_value) * factor
    except:
        logger.error("conversion failed for: %s" % str_value)
        duration = 0.0

    return duration


known_report_codes = {
    201:
    "Download successful. (variations: Downloaded, Inserted, Published, Copied, or Linked)",
    203: "Non-Authoritative Information: transformed during download.",
    205:
    "Reset Content: truncated. File is shorter than originally expected (changed length during transfer) This only arises during multi-part transfers.",
    205: "Reset Content: checksum recalculated on receipt.",
    304:
    "Not modified (Checksum validated, unchanged, so no download resulted.)",
    307: "Insertion deferred (writing to temporary part file for the moment.)",
    417: "Expectation Failed: invalid notification message (corrupt headers)",
    499: "Failure: Not Copied. SFTP/FTP/HTTP download problem",
    #FIXME : should  not have 503 error code 3 times in a row
    # 503: "Service unavailable. delete (File removal not currently supported.)",
    503: "Unable to process: Service unavailable",
    # 503: "Unsupported transport protocol specified in posting."
}


class Message(dict):
    """
        A notification message in Sarracenia is stored as a python dictionary, with a few extra management functions.

        The internal representation is very close to the v03 format defined here: https://metpx.github.io/sarracenia/Reference/sr_post.7.html

        Unfortunately, sub-classing of dict means that to copy it from a dict will mean losing the type,
        and hence the need for the copyDict member.
    """
    def __init__(self):
        self['_format'] = 'v03'
        self['_deleteOnPost'] = set(['_format'])


    def __computeIdentity(msg, path, o):
        """
           check extended attributes for a cached identity sum calculation.
           if extended attributes are present, and 
           * the file mtime is not too new, and 
           * the cached sum us using the same method
           then use the cached value.

           otherwise, calculate a checksum. 
           set the file's extended attributes for the new value.
           the method of checksum calculation is from options.identity.
           
           sets the message 'identity' field if appropriate.
        """
        xattr = sarracenia.filemetadata.FileMetadata(path)

        if o.randomize:
            methods = [
                'random', 'md5', 'md5name', 'sha512', 'cod,md5', 'cod,sha512'
            ]
            calc_method = random.choice(methods)
        elif 'identity' in xattr.x and 'mtime' in xattr.x:
            if xattr.get('mtime') >= msg['mtime']:
                logger.debug("mtime remembered by xattr")
                fxainteg = xattr.get('identity')
                if fxainteg['method'] == o.identity_method: 
                     msg['identity'] = fxainteg
                     return
                logger.debug("xattr different method than on disk")
                calc_method = o.identity
            else:
                logger.debug("xattr sum too old")
                calc_method = o.identity_method
        else:
            calc_method = o.identity_method

        if calc_method == None:
            return

        xattr.set('mtime', msg['mtime'])

        #logger.debug("sum set by compute_sumstr")

        if calc_method[:4] == 'cod,' and len(calc_method) > 2:
            sumstr = calc_method
        elif calc_method in [ 'md5name', 'invalid' ]:
            xattr.persist()  # persist the mtime, at least...
            return  # no checksum needed for md5name. 
        elif calc_method == 'arbitrary':
            sumstr = {
                'method': 'arbitrary',
                'value': o.identity_arbitrary_value
            }
        else:
            sumalgo = sarracenia.identity.Identity.factory(calc_method)
            sumalgo.set_path(path)

            # compute checksum

            if calc_method in ['md5', 'sha512']:

                fp = open(path, 'rb')
                i = 0
                while i < msg['size']:
                    buf = fp.read(o.bufsize)
                    if not buf: break
                    sumalgo.update(buf)
                    i += len(buf)
                fp.close()

            # setting sumstr
            checksum = sumalgo.value
            sumstr = {'method': calc_method, 'value': checksum}

        msg['identity'] = sumstr
        xattr.set('identity', sumstr)
        xattr.persist()

    def copyDict(msg, d):
        """
          copy dictionary into message.
       """
        if d is None: return

        for h in d:
            msg[h] = d[h]

    def dumps(msg) -> str:
        """
           FIXME: used to be msg_dumps.
           print a message in a compact but relatively compact way.
           msg is a python dictionary. if there is a field longer than maximum_field_length, 
           truncate.
    
       """

        maximum_field_length = 255

        if msg is None: return ""

        if msg['_format'] == 'Wis':
            s = '{ '
            if 'id' in msg:
                s += f"{{ 'id': '{msg['id']}', 'type':'Feature', "
            if 'geometry' in msg:
                s += f"'geometry':{msg['geometry']} 'properties':{{ "
            else:
                s += "'geometry': None, 'properties':{ "

        else:
            s = "{ "

        for k in sorted(msg.keys()):

            if msg['_format'] == 'v04' and k in [ 'id', 'type', 'geometry' ]:
               continue
            
            if type(msg[k]) is dict:
                if k != 'properties':
                    v = "{ "
                for kk in sorted(msg[k].keys()):
                    v += " '%s':'%s'," % (kk, msg[k][kk])
                v = v[:-1] 
                if k != 'properties':
                   v += " }"
            else:
                try:
                    v = "%s" % msg[k]
                except:
                    v = "unprintable"

            if len(v) > maximum_field_length:
                v = v[0:maximum_field_length - 4] + '...'
                if v[0] == '{':
                    v += '}'

            s += f" '{k}':'{v}'," 

        if msg['_format'] == 'Wis':
            s += ' } '

        s = s[:-1] + " }"
        return s

    @staticmethod
    def fromFileData(path, o, lstat=None):
        """
            create a message based on a given file, calculating the checksum.
            returns a well-formed message, or none.
        """
        m = sarracenia.Message.fromFileInfo(path, o, lstat)
        if lstat :
            if os_stat.S_ISREG(lstat.st_mode):
                m.__computeIdentity(path, o)
                if features['filetypes']['present']:
                    try:
                        t = magic.from_file(path,mime=True)
                        m['contentType'] = t
                    except Exception as ex:
                        logging.info("trying to determine mime-type. Exception details:", exc_info=True)
                #else:
                #    m['contentType'] = 'application/octet-stream' # https://www.rfc-editor.org/rfc/rfc2046.txt (default when clueless)
                # I think setting a bad value is worse than none, so just omitting.
            elif os_stat.S_ISDIR(lstat.st_mode):
                m['contentType'] = 'text/directory' # source: https://www.w3.org/2002/12/cal/rfc2425.html
            elif os_stat.S_ISLNK(lstat.st_mode):
                m['contentType'] = 'text/link' # I invented this one, could not find any reference
        return m

    @staticmethod
    def fromFileInfo(path, o, lstat=None):
        """
            based on the fiven information about the file (it's name and a stat record if available)
            and a configuration options object (sarracenia.config.Config) 
            return an sarracenia.Message suitable for placement on a worklist.

            A message is a specialized python dictionary with a certain set of fields in it.
            The message returned will have the necessary fields for processing and posting. 
    
            The message is built for a file is based on the given path, options (o), and lstat (output of os.stat)
             
            The lstat record is used to build 'atime', 'mtime' and 'mode' fields if 
            timeCopy and permCopy options are set.
    
            if no lstat record is supplied, then those fields will not be set.
        """

        msg = Message()

        #FIXME no variable substitution... o.variableExpansion ?
        if hasattr(o,'post_format') :
            msg['_format'] = o.post_format
        elif hasattr(o,'post_topicPrefix') and o.post_topicPrefix[0] in [ 'v02', 'v03' ]:
            msg['_format'] = o.post_topicPrefix[0]
        else:
            msg['_format'] = 'v03'

        if hasattr(o, 'post_exchange'):
            msg['exchange'] = o.post_exchange
        elif hasattr(o, 'exchange'):
            msg['exchange'] = o.exchange

        msg['local_offset'] = 0
        msg['_deleteOnPost'] = set(['exchange', 'local_offset', 'subtopic', '_format'])

        # notice
        msg['pubTime'] = timeflt2str(time.time())

        # set new_dir, new_file, new_subtopic, etc...
        msg.updatePaths(o, os.path.dirname(path), os.path.basename(path))

        # rename
        if 'new_relPath' in msg:
            post_relPath = msg['new_relPath']
        elif 'relPath' in msg:
            post_relPath = msg['relPath']
        else:
            post_relPath = None

        newname = post_relPath

        # rename path given with no filename

        if o.rename:
            msg['retrievePath'] = msg['new_retrievePath']
            newname = o.variableExpansion(o.rename)
            if o.rename[-1] == '/':
                newname += os.path.basename(path)

        # strip 'N' heading directories
        if o.strip > 0:
            strip = o.strip
            if path[0] == '/': strip = strip + 1
            # if we strip too much... keep the filename
            token = path.split('/')
            try:
                token = token[strip:]
            except:
                token = [os.path.basename(path)]
            newname = '/' + '/'.join(token)

        if newname != post_relPath: msg['rename'] = newname

        if hasattr(o, 'to_clusters') and (o.to_clusters is not None):
            msg['to_clusters'] = o.to_clusters
        if hasattr(o, 'cluster') and (o.cluster is not None):
            msg['from_cluster'] = o.cluster

        if hasattr(o, 'source') and (o.source is not None):
            msg['source'] = o.source


        if o.identity_method:
            if o.identity_method.startswith('cod,'):
                msg['identity'] = {
                    'method': 'cod',
                    'value': o.identity_method[4:]
                }
            elif o.identity_method in ['random']:
                algo = sarracenia.identity.Identity.factory(o.identity_method)
                algo.set_path(post_relPath)
                msg['identity'] = {
                    'method': o.identity_method,
                    'value': algo.value
                }
        else:
            if 'identity' in msg:
                   del msg['identity']
 
        # for md5name/aka None aka omit identity... should just fall through.

        if lstat is None: return msg

        if (lstat.st_mode is not None) :
            msg['mode'] = "%o" % (lstat.st_mode & 0o7777)
            if not o.permCopy:
                msg['_deleteOnPost'] |= set(['mode'])
            
            if os_stat.S_ISDIR(lstat.st_mode):
                msg['fileOp'] = { 'directory': '' }
                return msg

        if lstat.st_size is not None:
            msg['size'] = lstat.st_size

        if lstat.st_mtime is not None:
            msg['mtime'] = timeflt2str(lstat.st_mtime)
        if lstat.st_atime is not None:
            msg['atime'] = timeflt2str(lstat.st_atime)

        if not o.timeCopy:
            msg['_deleteOnPost'] |= set([ 'atime', 'mtime' ])

        return msg

    @staticmethod
    def fromStream(path, o, data=None):
        """
           Create a file and message for the given path.  
           The file will be created or overwritten with the provided data.
           then invoke fromFileData() for the resulting file.
        """

        with open(path, 'wb') as fh:
            fh.write(data)

        if hasattr(o, 'chmod') and o.chmod:
            os.chmod(path, o.chmod)

        return sarracenia.Message.fromFileData(path, o, stat(path))

    def setReport(msg, code, text=None):
        """
          FIXME: used to be msg_set_report
          set message fields to indicate result of action so reports can be generated.
    
          set is supposed to indicate final message dispositions, so in the case
          of putting a message on worklist.failed... no report is generated, since
          it will be retried later.  FIXME: should we publish an interim failure report?
    
        """

        if code in known_report_codes:
            if text is None:
                text = known_report_codes[code]
                
        else:
            logger.warning('unknown report code supplied: %d:%s' %
                           (code, text))
            if text is None:
                text = 'unknown disposition'

        if 'report' in msg:
            logger.warning('overriding initial report: %d: %s' %
                           (msg['report']['code'], msg['report']['message']))

        msg['report'] = {'code': code, 'timeCompleted': nowstr(), 'message': text}
        msg['_deleteOnPost'] |= set(['report'])

    def updatePaths(msg, options, new_dir=None, new_file=None):
        """
        set the new_* fields in the message based on changed file placement.
        if new_* options are ommitted updaste the rest of the fields in 
        the message based on their current values.

        If you change file placement in a flow callback, for example.
        One would change new_dir and new_file in the message.
        This routines updates other fields in the message (e.g. relPath, 
        baseUrl, topic ) to match new_dir/new_file.

        msg['post_baseUrl'] defaults to msg['baseUrl']
     
        """

        # the headers option is an override.
        if hasattr(options, 'fixed_headers'):
            for k in options.fixed_headers:
                msg[k] = options.fixed_headers[k]

        msg['_deleteOnPost'] |= set([
            'new_dir', 'new_file', 'new_relPath', 'new_baseUrl', 'new_subtopic', 'subtopic', 'post_format'
        ])
        if new_dir:
            msg['new_dir'] = new_dir

        if new_file or new_file == '':
            msg['new_file'] = new_file

        relPath = new_dir + '/' + new_file

        if options.post_baseUrl:
            baseUrl_str = options.variableExpansion(options.post_baseUrl, msg)
        else:
            if 'baseUrl' in msg:
                baseUrl_str = msg['baseUrl']
            else:
                logger.error('missing post_baseUrl setting')
                return

        if options.post_format:
            msg['post_format'] = options.post_format
        elif options.post_topicPrefix:
            msg['post_format'] = options.post_topicPrefix[0]
        elif options.topicPrefix != msg['_format']:
            logger.warning( f"received message in {msg['_format']} format, expected {options.post_topicPrefix} " )
            msg['post_format'] = options.topicPrefix[0]
        else:
            msg['post_format'] = msg['_format']
           
        if hasattr(options, 'post_baseDir') and ( type(options.post_baseDir) is str ) \
            and ( len(options.post_baseDir) > 1):
            pbd_str = options.variableExpansion(options.post_baseDir, msg)
            parsed_baseUrl = sarracenia.baseUrlParse(baseUrl_str)

            if relPath.startswith(pbd_str):
                relPath = new_dir.replace(pbd_str, '', 1) + '/' + new_file

            if (len(parsed_baseUrl.path) > 1) and relPath.startswith(
                    parsed_baseUrl.path):
                relPath = relPath.replace(parsed_baseUrl.path, '', 1)

        if ('new_dir' not in msg) and options.post_baseDir:
            msg['new_dir'] = options.post_baseDir
            
        msg['new_baseUrl'] = baseUrl_str

        if len(relPath) > 0 and relPath[0] == '/':
            relPath = relPath[1:]

        msg['new_relPath'] = relPath
        msg['new_subtopic'] = relPath.split('/')[0:-1]

        for i in ['relPath', 'subtopic', 'baseUrl']:
            if not i in msg:
                msg[i] = msg['new_%s' % i]

        if sys.platform == 'win32':
            if 'new_dir' not in msg:
                msg['new_dir'] = msg['new_dir'].replace('\\', '/')
            msg['new_relPath'] = msg['new_relPath'].replace('\\', '/')
            if re.match('[A-Z]:', str(options.currentDir),
                        flags=re.IGNORECASE):
                msg['new_dir'] = msg['new_dir'].lstrip('/')
                msg['new_relPath'] = msg['new_relPath'].lstrip('/')

    def validate(msg):
        """
        FIXME: used to be msg_validate
        return True if message format seems ok, return True, else return False, log some reasons.
        """
        if not type(msg) is sarracenia.Message:
            return False

        res = True
        for required_key in ['pubTime', 'baseUrl', 'relPath']:
            if not required_key in msg:
                logger.error('missing key: %s' % required_key)
                res = False

        if not timeValidate(msg['pubTime']):
            res = False

        if not res:
            logger.error('malformed message: %s', msg)
        return res

    def getContent(msg):
        """
           Retrieve the data referred to by a message.  The data may be embedded
           in the messate, or this routine may resolve a link to an external server 
           and download the data.

           does not handle authentication.
           This routine is meant to be used with small files. using it to download
           large files may be very inefficient. Untested in that use-case.

           Return value is the data.
        """

        # inlined/embedded case.
        if 'content' in msg:
            if msg['content']['encoding'] == 'base64':
                return b64decode(msg['content']['value'])
            else:
                return msg['content']['value'].encode('utf-8')
        # case requiring resolution.
        if 'retrievePath' in msg:
            retUrl = msg['baseUrl'] + '/' + msg['retrievePath']
        else:
            retUrl = msg['baseUrl'] + '/' + msg['relPath']

        with urllib.request.urlopen(retUrl) as response:
            return response.read()




"""
Plugin downloadbaseurl.py:
    Downloads files sourced from the baseUrl of the poster, and saves them in the
    directory specified in the config. Created to use with the poll_nexrad.py
    plugin to download files uploaded in the NEXRAD US Weather Radar public dataset.
    Compatible with Python 3.5+.

Example:
    A sample do_download option for subscribe.
	Downloads the file located at message['baseUrl'] and saves it

Usage:
	flowcb sarracenia.flowcb.accept.downloadbaseurl.DownloadBaseUrl
"""

import logging
import os
import urllib.request
from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)

class DownloadBaseUrl(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            # if mirror is set to True, comment these two lines out
            #TODO: this self.o.new_dir could be instead message['new_dir'] I think.. to see..
            keypath, key = os.path.split(self.o.new_dir + message['new_file'])
            if not os.path.exists(keypath): os.makedirs(keypath)

            with open(keypath + '/' + key, 'wb') as f:
                with urllib.request.urlopen(message['baseUrl']) as k:
                    f.write(k.read())



import logging
import paramiko
import sarracenia
from sarracenia import nowflt, timestr2flt
from sarracenia.flowcb.poll import Poll

logger = logging.getLogger(__name__)


class Nasa_mls_nrt(Poll):
    def handle_data(self, data):
        """
           decode some HTML into an SFTPAttributes record for a file.
        """

        st = paramiko.SFTPAttributes()
        st.st_mtime = 0
        st.st_mode = 0o775
        st.filename = data

        if 'MLS-Aura' in data:
            logger.debug("data %s" % data)
            self.entries[data] = st

            logger.info("(%s) = %s" % (self.myfname, st))
        if self.myfname == None: return
        if self.myfname == data: return



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# Questions or bugs report: dps-client@ec.gc.ca
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
# credentials.py : python3 utility tool to configure all protocol credentials
#
#
# Code contributed by:
#  Michel Grenier - Shared Services Canada
#  Last Changed   : Dec 29 11:42:11 EST 2015
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#

#

import logging

logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)

import os
import re
import urllib, urllib.parse
import sys


class Credential:
    """An object that holds information about a credential, read from a 
    credential file, which has one credential per line, format::
        url option1=value1, option2=value2
       
    Examples::
        sftp://alice@herhost/ ssh_keyfile=/home/myself/mykeys/.ssh.id_dsa
        ftp://georges:Gpass@hishost/  passive = True, binary = True
       
    `Format Documentation. <https://metpx.github.io/sarracenia/Reference/sr3_credentials.7.html>`_

    Attributes:
        url (urllib.parse.ParseResult): object with URL, password, etc.
        ssh_keyfile (str): path to SSH key file for SFTP
        passive (bool): use passive FTP mode, defaults to ``True``
        binary (bool): use binary FTP mode, defaults to ``True``
        tls (bool): use FTPS with TLS, defaults to ``False``
        prot_p (bool): use a secure data connection for TLS
        bearer_token (str): bearer token for HTTP authentication
        login_method (str): force a specific login method for AMQP (PLAIN,
            AMQPLAIN, EXTERNAL or GSSAPI)

    Usage:

       # build a credential from a url string:

       from sarracenia.credentials import Credential

       broker = Credential('amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')



    """

    def __init__(self, urlstr=None):
        """Create a Credential object.

            Args:
                urlstr (str): a URL in string form to be parsed.
        """

        if urlstr is not None:
            self.url = urllib.parse.urlparse(urlstr)
        else:
            self.url = None

        self.ssh_keyfile = None
        self.passive = True
        self.binary = True
        self.tls = False
        self.prot_p = False
        self.bearer_token = None
        self.login_method = None

    def __str__(self):
        """Returns attributes of the Credential object as a readable string.
        """

        s = ''
        if False:
            s += self.url.geturl()
        else:
            s += self.url.scheme + '://'
            if self.url.username:
                s += self.url.username
            #if self.url.password:
            #   s += ':' + self.url.password
            if self.url.hostname:
                s += '@' + self.url.hostname
            if self.url.port:
                s += ':' + str(self.url.port)
            if self.url.path:
                s += self.url.path

        s += " %s" % self.ssh_keyfile
        s += " %s" % self.passive
        s += " %s" % self.binary
        s += " %s" % self.tls
        s += " %s" % self.prot_p
        s += " %s" % self.bearer_token
        s += " %s" % self.login_method
        return s


class CredentialDB:
    """Parses, stores and manages Credential objects.

    Attributes:
        credentials (dict): contains all sarracenia.credentials.Credential objects managed by the CredentialDB.

    Usage:
       # build a credential via lookup in the normal files: 
       import CredentialDB from sarracenia.credentials

       credentials = CredentialDB.read( "/the/path/to/the/credentials.conf" )

       # if there are corresponding passwords or modulation of login information look it up.
       
       broker = credentials.get( "amqps://hpfx.collab.science.gc.ca" )
       remote = credentials.get( "sftp://hoho@theserver" )
    """

    def __init__(self, Unused_logger=None):
        """Create a CredentialDB.

        Args:
            Unused_logger: logger argument no longer used... left there for API compat with old calls.
        """
        self.credentials = {}
        self.pwre = re.compile(':[^/:]*@')

        logger.debug("__init__")

    def add(self, urlstr, details=None):
        """Add a new credential to the DB.

        Args:
            urlstr (str): string-formatted URL to be parsed and added to DB.
            details (sarracenia.credentials.Credential): a Credential object can be passed in, otherwise one is
                created by parsing urlstr.
        """

        # need to create url object
        if details == None:
            details = Credential()
            details.url = urllib.parse.urlparse(urlstr)

        self.credentials[urlstr] = details

    def get(self, urlstr):
        """Retrieve a Credential from the DB by urlstr. If the Credential is valid, but not already cached, it will be
        added to the CredentialDB.

        Args:
            urlstr (str): credentials as URL string to be parsed.
        
        Returns:
            tuple: containing
                cache_result (bool): ``True`` if the credential was retrieved from the CredentialDB cache, ``False``
                    if it was not in the cache. Note that ``False`` does not imply the Credential or urlstr is
                    invalid.
                credential (sarracenia.credentials.Credential): the Credential
                    object matching the urlstr, ``None`` if urlstr is invalid.
        """
        #logger.debug("CredentialDB get %s" % urlstr)

        # already cached

        if self.has(urlstr):
            #logger.debug("CredentialDB get in cache %s %s" % (urlstr,self.credentials[urlstr]))
            return True, self.credentials[urlstr]

        # create url object if needed

        url = urllib.parse.urlparse(urlstr)

        # add anonymous default, if necessary.
        if ( 'amqp' in url.scheme ) and \
           ( (url.username == None) or (url.username == '') ):
            urlstr = urllib.parse.urlunparse( ( url.scheme, \
                'anonymous:anonymous@%s' % url.netloc, url.path, None, None, url.port ) )
            url = urllib.parse.urlparse(urlstr)
            if self.isValid(url):
                self.add(urlstr)
                return False, self.credentials[urlstr]

        # resolved from defined credentials

        ok, details = self._resolve(urlstr, url)
        if ok: return True, details

        # not found... is it valid ?
        if not self.isValid(url):
            return False, None

        # cache it as is... we dont want to validate every time

        self.add(urlstr)
        return False, self.credentials[urlstr]

    def has(self, urlstr):
        """Return ``True`` if the Credential matching the urlstr is already in the CredentialDB.
        
        Args:
            urlstr(str): credentials in a URL string.
        """
        logger.debug("has %s" % urlstr)
        return urlstr in self.credentials

    def isTrue(self, S):
        """Returns ``True`` if s is ``true``, ``yes``, ``on`` or ``1``.

        Args:
            S (str): string to check if true.
        """
        s = S.lower()
        if s == 'true' or s == 'yes' or s == 'on' or s == '1': return True
        return False

    def isValid(self, url, details=None):
        """Validates a URL and Credential object. Checks for empty passwords, schemes, etc.
            
        Args:
            url (urllib.parse.ParseResult): ParseResult object for a URL.
            details (sarracenia.credentials.Credential): sarra Credential object containing additional details about
                the URL.

        Returns:
            bool: ``True`` if a URL is valid, ``False`` if not.
        """

        # network location
        if url.netloc == '':
            # file (why here? anyway)
            if url.scheme == 'file': return True
            logger.error( f'no network location, and not a file url' )
            return False

        # amqp... vhost not check: default /

        # user and password provided we are ok
        user = url.username != None and url.username != ''
        pasw = url.password != None and url.password != ''
        both = user and pasw

        # we have everything
        if both: return True

        # we have no user and no pasw (http normal, https... no cert,  sftp hope for .ssh/config)
        if not user and not pasw:
            if url.scheme in ['http', 'https', 'sftp']: return True
            logger.error( f'unknown scheme: {url.scheme}')
            return False

        #  we have a pasw no user
        if pasw:
            # not sure... sftp hope to get user from .ssh/config
            if url.scheme == 'sftp': return True
            logger.error( f'password with no username specified')
            return False

        #  we only have a user ... permitted only for sftp

        if url.scheme != 'sftp': 
            logger.error( f'credential not found'  )
            return False

        #  sftp and an ssh_keyfile was provided... check that it exists

        if details and details.ssh_keyfile:
            if not os.path.exists(details.ssh_keyfile): 
                logger.error( f'ssh_keyfile not found: {details.ssh_keyfile}')
                return False

        #  sftp with a user (and perhaps a valid ssh_keyfile)

        return True

    def _parse(self, line):
        """Parse a line of a credentials file, add it to the CredentialDB.

        Args:
            line (str): line to be parsed.
        """
        #logger.debug("parse %s" % self.pwre.sub(':<secret!>@', line, count=1) )

        try:
            sline = line.strip()
            if len(sline) == 0 or sline[0] == '#': return

            # first field url string = protocol://user:password@host:port[/vost]
            parts = sline.split()
            urlstr = parts[0]
            url = urllib.parse.urlparse(urlstr)

            # credential details
            details = Credential()
            details.url = url

            # no option
            if len(parts) == 1:
                if not self.isValid(url, details):
                    logger.error("bad credential 1 (%s)" % line)
                    return
                self.add(urlstr, details)
                return

            # parsing options :  comma separated option names
            # some option has name = value : like ssh_keyfile

            optline = sline.replace(urlstr, '')
            optline = optline.strip()
            optlist = optline.split(',')

            for optval in optlist:
                parts = optval.split('=')
                keyword = parts[0].strip()

                if keyword == 'ssh_keyfile':
                    details.ssh_keyfile = os.path.expandvars(os.path.expanduser(parts[1].strip()))
                elif keyword == 'passive':
                    details.passive = True
                elif keyword == 'active':
                    details.passive = False
                elif keyword == 'binary':
                    details.binary = True
                elif keyword == 'ascii':
                    details.binary = False
                elif keyword == 'ssl':
                    details.tls = False
                elif keyword == 'tls':
                    details.tls = True
                elif keyword == 'prot_p':
                    details.prot_p = True
                elif keyword in ['bearer_token', 'bt']:
                    details.bearer_token = parts[1].strip()
                elif keyword == 'login_method':
                    details.login_method = parts[1].strip()
                else:
                    logger.warning("bad credential option (%s)" % keyword)

            # need to check validity
            if not self.isValid(url, details):
                logger.error("bad credential 2 (%s)" % line)
                return

            # seting options to protocol

            self.add(urlstr, details)

        except:
            logger.error("credentials/parse %s" % line)
            logger.debug('Exception details: ', exc_info=True)

    def read(self, path):
        """Read in a file containing credentials (e.g. credentials.conf). All credentials are parsed and added to the
        CredentialDB.

        Args:
            path (str): path of file to be read.
        """
        logger.debug("read")

        # read in provided credentials (not mandatory)
        try:
            if os.path.exists(path):
                with open(path) as f:
                    lines = f.readlines()

                for line in lines:
                    self._parse(line)
        except:
            logger.error("credentials/read path = %s" % path)
            logger.debug('Exception details: ', exc_info=True)
        #logger.debug("Credentials = %s\n" % self.credentials)

    def _resolve(self, urlstr, url=None):
        """Resolve credentials for AMQP vhost from ones passed as a string, and optionally a urllib.parse.ParseResult
        object, into a sarracenia.credentials.Credential object.

        Args:
            urlstr (str): credentials in a URL string.
            url (urllib.parse.ParseResult): ParseResult object with creds.

        Returns:
            tuple: containing
                result (bool): ``False`` if the creds were not in the CredentialDB. ``True`` if they were.
                details (sarracenia.credentials.Credential): the updated Credential object, or ``None``.
                

        """

        # create url object if needed

        if not url:
            url = urllib.parse.urlparse(urlstr)

        # resolving credentials

        for s in self.credentials:
            details = self.credentials[s]
            u = details.url

            if url.scheme != u.scheme: continue
            if url.hostname != u.hostname: continue
            if url.port != u.port: continue
            if url.username != u.username:
                if url.username != None: continue
            if url.password != u.password:
                if url.password != None: continue

            # for AMQP...  vhost checking
            # amqp users have same credentials for any vhost
            # default /  may not be set...
            if 'amqp' in url.scheme:
                url_vhost = url.path
                u_vhost = u.path
                if url_vhost == '': url_vhost = '/'
                if u_vhost == '': u_vhost = '/'

                if url_vhost != u_vhost: continue

            # resolved : cache it and return

            self.credentials[urlstr] = details
            #logger.debug("Credentials get resolved %s %s" % (urlstr,details))
            return True, details

        return False, None



"""
This plugin prints a traceback when the program is stopped

  plugin trace_on_stop

"""
import io
import logging
import traceback
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class TRACE_ON_STOP(FlowCB):
    def __init__(self):
        pass

    def on_stop(self):
        self.LOG_TRACE()
        return True

    def LOG_TRACE(self):
        tb_output = io.StringIO()
        traceback.print_stack(None, None, tb_output)
        logger.info("\n\n****************************************\n" + \
                             "***** PRINTING TRACEBACK FROM STOP *****\n" + \
                             "****************************************\n" + \
                           "\n" + tb_output.getvalue()             + "\n" + \
                           "\n****************************************\n")
        tb_output.close()



import copy
import json
import logging
import sys
import time
from sarracenia.featuredetection import features

import sarracenia

logger = logging.getLogger(__name__)

default_options = {
    'acceptUnmatched': True,
    'batch': 100,
    'bindings': [],
    'broker': None,
    'dry_run': False,
    'exchange': 'xpublic',
    'expire': 300,
    'inline': False,
    'inlineEncoding': 'guess',
    'inlineByteMax': 4096,
    'logFormat':
    '%(asctime)s [%(levelname)s] %(name)s %(funcName)s %(message)s',
    'logLevel': 'info',
    'messageDebugDump': False,
    'message_strategy': {
        'reset': True,
        'stubborn': True,
        'failure_duration': '5m'
    },
    'message_ttl': 0,
    'topicPrefix': ['v03'],
    'tlsRigour': 'normal'
}

def ProtocolPresent(p) -> bool:
    if ( p[0:4] in ['amqp'] ) and sarracenia.features['amqp']['present']:
       return True
    if ( p[0:4] in ['mqtt'] ) and sarracenia.features['mqtt']['present']:
       return True
    if p in sarracenia.features:
        logger.critical( f"support for {p} missing, please install python packages: {' '.join(sarracenia.features[p]['modules_needed'])}" )
    else:
        logger.critical( f"Protocol scheme {p} unsupported for communications with message brokers" )

    sys.exit(1)
    return False


class Moth():
    """
        Moth ... Messages Organized by Topic Headers
        (en français: Messages organisés par thème hierarchique. )

        A multi-protocol library for use by hierarchical message passing implementations,
        (messages which have a 'topic' header that is used for routing by brokers.)
 
        - regardless of protocol, the message format returned should be the same.
        - the message is turned into a sarracenia.Message object, which acts like a python 
          dictionary, corresponding to key-value pairs in the message body, and properties.
        - topic is a special key that may end up in the message body, or some sort of property
          or metadata.
        - the protocol should support acknowledgement under user control. Such control indicated
          by the presence of an entry_point called "ack". The entry point accepts "ack_id" as
          a message identifier to be passed to the broker. Whatever protocol symbol is used
          by the protocol, it is passed through this message property. Examples:
          in rabbitmq/amqp ack takes a "delivery_tag" as an argument, in MQTT, it takes a "message-id"
          so when receiving an AMQP message, the m['ack_id'] is assigned the delivery_tag from the message. 
        - There is a special dict item:  "_DeleteOnPost",  
          to identify keys which are added only for local use.
          they will be removed from the message when publishing.
          examples:  topic (sent outside body), message-id (used for acknowledgements.)
          new_basedir, ack_id, new\_... (settings...)

        Intent is to be specialized for topic based data distribution (MQTT style.)
        API to allow pass-through of protocol specific properties, but apply templates for genericity.

        Target protocols (and corresponding libraries.): AMQP, MQTT, ?

        Things to specify:

        * broker

        * topicPrefix

        * subTopic  

        * queueName   (for amqp, used as client-id for mqtt)

        this library knows nothing about Sarracenia, the only code used from sarracenia is to interpret
        duration properties, from the root sarracenia/__init__.py, the broker argument from sarracenia.credentials
  
        usage::

           import sarracenia.moth
           import sarracenia.credentials


           props = sarracenia.moth.default_options
           props['broker'] = sarracenia.credentials.Credential('amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')
           props['expire'] = 300
           props['batch'] = 1
           is_subscriber=True

           c= Moth( props, is_subscriber  )

           messages = c.newMessages()

           # if there are new messages from a publisher, return them, otherwise return
           # an empty list []].
             
           p=Moth( { 'batch':1 }, False )

           p.putNewMessage()

           p.close()
           # tear down connection.     
  
        Initialize a broker connection. Connections are unidirectional.
        either for subscribing (with subFactory) or publishing (with pubFactory.)
       
        The factories return objects subclassed to match the protocol required
        by the broker argument.

        arguments to the factories are:

        * broker ... the url of the broker to connect to.

        * props is a dictionary or properties/parameters.

        * supplied as overrides to the default properties listed above.

        Some may vary among protocols::
 
          Protocol     library implementing    URL to select
          --------     --------------------    -------------

          AMQPv0.9 --> amqplib from Celery --> amqp, amqps

          AMQPv0.9 --> pika                --> pika, pikas

          MQTTv3   --> paho                --> mqtt, mqtts

          AMQPv1.0 --> qpid-proton         --> amq1, amq1s



       **messaging_strategy**

       how to manage the connection. Covers whether to treat the connection
       as new or assume it is set up. Also, If something goes wrong.  
       What should be done. 
         
       * reset: on startup... erase any state, and re-initialize.

       * stubborn: If set to True, loop forever if something bad happens.  
         Never give up. This sort of setting is desired in operations, especially unattended.
         if set to False, may give up more easily.

       * failure_duration is to advise library how to structure connection service level.
          
         * 5m - make a connection that will recover from transient errors of a few minutes,
           but not tax the broker too much for prolonged outages.

         * 5d - duration outage to striving to survive connection for five days.

       Changing recovery_strategy setting, might result in having to destroy and re-create 
       consumer queues (AMQP.)

       **Options**

       **both**

       * 'topicPrefix' : [ 'v03' ]

       * 'messageDebugDump': False, --> enable printing of raw messages.

       * 'inline': False,  - Are we inlining content within messages?

       * 'inlineEncoding': 'guess', - what encoding should we use for inlined content?

       * 'inlineByteMax': 4096,  - Maximum size of messages to inline.

       **for get**

       *  'batch'  : 100  # how many messages to get at once

       *  'broker' : an sr_broker ?

       *  'queueName'  : Mandatory, name of a queue. (only in AMQP... hmm...)

       *  'bindings' : [ list of bindings ]

       *  'loop'

       **optional:**

       *  'message_ttl'    

       **for put:**

       *   'exchange' (only in AMQP... hmm...)
       

    """
    @staticmethod
    def subFactory(props) -> 'Moth':

        if not props['broker'] :
            logger.error('no broker specified')
            return None

        if not hasattr(props['broker'],'url'):
            logger.error('invalid broker url')
            return None

        if not ProtocolPresent(props['broker'].url.scheme):
           logger.error('unknown broker scheme/protocol specified')
           return None

        for sc in Moth.findAllSubclasses(Moth):
            driver=sc.__name__.lower()
            # when amqp_consumer option is True, use the moth.AMQPConsumer class, not normal moth.AMQP
            if "amqp_consumer" in props and props["amqp_consumer"]:
                if driver == 'amqp':
                    continue
                if driver == 'amqpconsumer':
                    # driver needs to be amqp to match with the broker URL's scheme
                    driver = 'amqp'
            scheme=props['broker'].url.scheme
            if (scheme == driver) or \
               ( (scheme[0:-1] == driver) and (scheme[-1] in [ 's', 'w' ])) or \
               ( (scheme[0:-2] == driver) and (scheme[-2] == 'ws')):
                return sc(props, True)
        logger.error('broker intialization failure')
        return None

    @staticmethod
    def pubFactory(props) -> 'Moth':
        if not props['broker']:
            logger.error('no broker specified')
            return None

        if not hasattr(props['broker'],'url'):
            logger.error('invalid broker url')
            return None

        if not ProtocolPresent(props['broker'].url.scheme):
           logger.error('unknown broker scheme/protocol specified')
           return None

        for sc in Moth.__subclasses__():
            driver=sc.__name__.lower()
            scheme=props['broker'].url.scheme
            if (scheme == driver) or \
               ( (scheme[0:-1] == driver) and (scheme[-1] in [ 's', 'w' ])) or \
               ( (scheme[0:-2] == driver) and (scheme[-2] == 'ws')):
                return sc(props, False)

        # ProtocolPresent test should ensure that we never get here...
        logger.error('broker intialization failure')
        return None
    
    @staticmethod
    def findAllSubclasses(cls) -> set:
        """Recursively finds all subclasses of a class. __subclasses__() only gives direct subclasses.
        """
        cls_subclasses = set(cls.__subclasses__())
        for sc in cls_subclasses:
            cls_subclasses = cls_subclasses.union(Moth.findAllSubclasses(sc))
        return cls_subclasses

    def __init__(self, props=None, is_subscriber=True) -> None:
        """
           If is_subscriber=True, then this is a consuming instance.
           expect calls to get* routines.

           if is_subscriber=False, then expect/permit only calls to put*
        """

        self.is_subscriber = is_subscriber
        self.connected=False
        self.please_stop = False
        self.metrics = { 'connected': False }
        self.metricsReset()

        if (sys.version_info.major == 3) and (sys.version_info.minor < 7):
            self.o = {}
            for k in default_options:
                if k == 'masks':
                    self.o[k] = default_options[k]
                else:
                    self.o[k] = copy.deepcopy(default_options[k])
        else:
            self.o = copy.deepcopy(default_options)

        if props is not None:
            self.o.update(props)

        me = 'sarracenia.moth.Moth'

        # apply settings from props.
        if 'settings' in self.o:
            if me in self.o['settings']:
                for s in self.o['settings'][me]:
                    self.o[s] = self.o['settings'][me][s]

        logging.basicConfig(format=self.o['logFormat'],
                            level=getattr(logging, self.o['logLevel'].upper()))

    def ack(self, message: sarracenia.Message ) -> None:
        """
          tell broker that a given message has been received.

          ack uses the 'ack_id' property to send an acknowledgement back to the broker.
        """
        logger.error("ack unimplemented")

    @property
    def default_options(self) -> dict:
        """
        get default properties to override, used by client for validation. 

        """
        return Moth.__default_options

    def getNewMessage(self) -> sarracenia.Message:
        """
        If there is one new message available, return it. Otherwise return None. Do not block.

        side effects:
            metrics.
            self.metrics['RxByteCount'] should be incremented by size of payload.
            self.metrics['RxGoodCount'] should be incremented by 1 if a good message is received.
            self.metrics['RxBadCount'] should be incremented by 1 if an invalid message is received (&discarded.)
        """
        logger.error("getNewMessage unimplemented")
        return None

    def newMessages(self) -> list:
        """
        If there are new messages available from the broker, return them, otherwise return None.

        On Success, this routine returns immediately (non-blocking) with either None, or a list of messages.

        On failure, this routine blocks, and loops reconnecting to broker, until interaction with broker is successful.
        
        """
        logger.error("NewMessages unimplemented")
        return []

    def putNewMessage(self, message:sarracenia.Message, content_type: str ='application/json') -> bool:
        """
           publish a message as set up to the given topic.

           return True is succeeded, False otherwise.

           side effect
            self.metrics['TxByteCount'] should be incremented by size of payload.
            self.metrics['TxGoodCount'] should be incremented by 1 if a good message is received.
            self.metrics['TxBadCount'] should be incremented by 1 if an invalid message is received (&discarded.)
        """
        logger.error("implementation missing!")
        return False

    def metricsReset(self) -> None:
        self.metrics['disconnectLast'] = 0
        self.metrics['disconnectTime'] = 0
        self.metrics['disconnectCount'] = 0
        self.metrics['rxByteCount'] = 0
        self.metrics['rxGoodCount'] = 0
        self.metrics['rxBadCount'] = 0
        self.metrics['txByteCount'] = 0
        self.metrics['txGoodCount'] = 0
        self.metrics['txBadCount'] = 0

    def metricsReport(self) -> tuple:
        if not self.metrics['connected'] and (self.metrics['disconnectLast'] > 0):
            down_time = time.time() - self.metrics['disconnectLast']
            self.metrics['disconnectTime'] += down_time

        return self.metrics

    def metricsConnect(self) -> None:
        self.metrics['connected']=True
        if self.metrics['disconnectLast'] > 0 :
            down_time = time.time() - self.metrics['disconnectLast']
            self.metrics['disconnectTime'] += down_time

    def metricsDisconnect(self) -> None:
        """
           tear down an existing connection.
        """
        self.metrics['connected']=False
        self.metrics['disconnectCount'] += 1
        self.metrics['disconnectLast'] = time.time()

    def close(self) -> None:
        logger.error("implementation missing!")

    def cleanup(self) -> None:
        """
          get rid of server-side resources associated with a client. (queues/id's, etc...)
        """
        if self.is_subscriber:
            self.getCleanUp()
        else:
            self.putCleanUp()

if features['amqp']['present']:
    import sarracenia.moth.amqp
    import sarracenia.moth.amqpconsumer

if features['mqtt']['present']:
    import sarracenia.moth.mqtt



"""
Plugin printlag.py:
     This is an after_accept plugin.
     print a message indicating how old messages received are.
     this should be used as an on_part script. For each part received it will print a line
     in the local log that looks like this:

     2015-12-23 22:54:30,328 [INFO] posted: 20151224035429.115 (lag: 1.21364 seconds ago) to deliver: /home/peter/test/dd/bulletins/alphanumeric/20151224/SA/EGGY/03/SAUK32_EGGY_240350__EGAA_64042,

     the number printed after "lag:" the time between the moment the message was originally posted on the server,
     and the time the script was called, which is very near the end of writing the file to local disk.

     This can be used to gauge whether the number of instances or internet link are sufficient
     to transfer the data selected.  if the lag keeps increasing, then something must be done.

Usage:
    flowcb sarracenia.flowcb.accept.printlag.PrintLag

"""
import calendar
import logging
import os
import stat
import time
from sarracenia import timestr2flt, nowflt
from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class PrintLag(FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            then = timestr2flt(message['pubTime'])
            now = nowflt()

            logger.info("print_lag, posted: %s, lag: %.2f sec. to deliver: %s, " % (message['pubTime'], (now - then), message['new_file']))
